/************************************************************************
 *   IRC - Internet Relay Chat, include/struct.h
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __struct_include__
#define __struct_include__

#include <stdio.h>
#include <sys/types.h>
#ifndef _WIN32
# include <netinet/in.h>
# include <netdb.h>
#else
# include <windows.h>
#endif

#include "config.h"
#include "common.h"

typedef struct SqlineItem  aSqlineItem;
typedef struct ConfItem    aConfItem;
typedef struct Notify      aNotify;
typedef struct Client      aClient;
typedef struct Channel     aChannel;
typedef struct User        anUser;
typedef struct Server      aServer;
typedef struct SLink       Link;
typedef struct SBan        Ban;
typedef struct SMode       Mode;
typedef struct ListOptions LOpts;

typedef struct CloneItem   aClone;

#ifdef NEED_U_INT32_T
typedef unsigned int u_int32_t; /* XXX Hope this works! */
#endif

#ifndef VMSP
# include "class.h"
# include "dbuf.h" /* THIS REALLY SHOULDN'T BE HERE!!! --msa */
#endif

#define MAXMSGLEN 2048 /* real message size limit */

#define        HOSTLEN         63 /* Length of hostname.  Updated to         */
                                  /* comply with RFC1123                     */

#define        NICKLEN         30 /* Necessary to put 9 here instead of 10
                                  ** if s_msg.c/m_nick has been corrected.
                                  ** This preserves compatibility with old
                                  ** servers --msa
                                  */
#define        USERLEN         10
#define        REALLEN         100
#define        TOPICLEN        307
/* DAL MADE ME PUT THIS IN THE FIEND:  --Russell
 * This number will be expanded to 200 in the near future
 */
#define        CHANNELLEN      32
#define        PASSWDLEN       32      /* orig. 20, changed to 32 for nickpasswords */
#define        KEYLEN          23
#define        BUFSIZE         512             /* WARNING: *DONT* CHANGE THIS!!!! */
#define        MAXRECIPIENTS   20
#define        MAXKILLS        20
#define        MAXBANS         60
#define        MAXBANLENGTH    1024
#define        MAXSILES        5
#define        MAXSILELENGTH   128
/*
 * Watch it - Don't change this unless you also change the ERR_TOOMANYWATCH
 * and PROTOCOL_SUPPORTED settings.
 */
#define MAXWATCH       128

#define        USERHOST_REPLYLEN       (NICKLEN+HOSTLEN+USERLEN+5)

/* NOTE: this must be down here so the stuff from struct.h IT uses works */
#include "whowas.h"


/*
** 'offsetof' is defined in ANSI-C. The following definition
** is not absolutely portable (I have been told), but so far
** it has worked on all machines I have needed it. The type
** should be size_t but...  --msa
*/
#ifndef offsetof
# define offsetof(t,m) (int)((&((t *)0L)->m))
#endif

#define elementsof(x) (sizeof(x)/sizeof(x[0]))

/* flags for bootup options (command line flags) */
#define BOOT_CONSOLE	0x00000001
#define BOOT_QUICK	0x00000002
#define BOOT_DEBUG	0x00000004
#define BOOT_INETD	0x00000008
#define BOOT_TTY	0x00000010
#define BOOT_OPER	0x00000020
#define BOOT_AUTODIE	0x00000040
#define BOOT_STDERR	0x00000080

#define STAT_LOG        -6      /* logfile for -x */
#define STAT_MASTER     -5      /* Local ircd master before identification */
#define STAT_CONNECTING -4
#define STAT_HANDSHAKE  -3
#define STAT_ME         -2
#define STAT_UNKNOWN    -1
#define STAT_SERVER     0
#define STAT_CLIENT     1
#define STAT_SERVICE    2       /* Services not implemented yet */

/* status macros. */

#define IsRegisteredUser(x) ((x)->status == STAT_CLIENT)
#define IsRegistered(x)     ((x)->status >= STAT_SERVER)
#define IsConnecting(x)     ((x)->status == STAT_CONNECTING)
#define IsHandshake(x)      ((x)->status == STAT_HANDSHAKE)
#define IsMe(x)             ((x)->status == STAT_ME)
#define IsUnknown(x)        ((x)->status == STAT_UNKNOWN || (x)->status == STAT_MASTER)
#define IsServer(x)         ((x)->status == STAT_SERVER)
#define IsClient(x)         ((x)->status == STAT_CLIENT)
#define IsLog(x)            ((x)->status == STAT_LOG)
#define IsService(x)        0

#define SetMaster(x)      ((x)->status = STAT_MASTER)
#define SetConnecting(x)  ((x)->status = STAT_CONNECTING)
#define SetHandshake(x)   ((x)->status = STAT_HANDSHAKE)
#define SetMe(x)          ((x)->status = STAT_ME)
#define SetUnknown(x)     ((x)->status = STAT_UNKNOWN)
#define SetServer(x)      ((x)->status = STAT_SERVER)
#define SetClient(x)      ((x)->status = STAT_CLIENT)
#define SetLog(x)         ((x)->status = STAT_LOG)
#define SetService(x)     ((x)->status = STAT_SERVICE)

#define FLAGS_PINGSENT    0x00000001 /* Unreplied ping sent */
#define FLAGS_DEADSOCKET  0x00000002 /* Local socket is dead--Exiting soon */
#define FLAGS_KILLED      0x00000004 /* Prevents "QUIT" from being sent for this */
#define FLAGS_BLOCKED     0x00000008 /* socket is in a blocked condition */
#define FLAGS_CLOSING     0x00000010 /* set when closing to suppress errors */
#define FLAGS_LISTEN      0x00000020 /* used to mark clients which we listen() on */
#define FLAGS_CHKACCESS   0x00000040 /* ok to check clients access if set */
#define FLAGS_DOINGDNS    0x00000080 /* client is waiting for a DNS response */
#define FLAGS_AUTH        0x00000100 /* client is waiting on rfc931 response */
#define FLAGS_WRAUTH      0x00000200 /* set if we havent writen to ident server */
#define FLAGS_LOCAL       0x00000400 /* set for local clients */
#define FLAGS_GOTID       0x00000800 /* successful ident lookup achieved */
#define FLAGS_DOID        0x00001000 /* I-lines say must use ident return */
#define FLAGS_NONL        0x00002000 /* No \n in buffer */
#define FLAGS_TS8         0x00004000 /* Why do you want to know? */
#define FLAGS_ULINE       0x00008000 /* User/server is considered U-lined */
#define FLAGS_SQUIT       0x00010000 /* Server has been /squit by an oper */
#define FLAGS_PROTOCTL    0x00020000 /* Received a PROTOCTL message */
#define FLAGS_PASSED      0x00040000 /* Client passed RegisterUser() */
#define FLAGS_MAP         0x00080000 /* Show this entry in /map */
#define FLAGS_ZIPPED_IN   0x00100000 /* incoming traffic for this link is zipped */
#define FLAGS_ZIPPED_OUT  0x00200000 /* outgoing traffic for this link is zipped */
#define FLAGS_IN_BURST    0x00400000 /* receiving their data */
#define FLAGS_OUT_BURST   0x00800000 /* sending my data */


/* Dec 26th, 1997 - added flags2 when I ran out of room in flags -DuffJ
 *
 * Dec 26th, 1997 - having a go at
 * splitting flags into flags and umodes
 * -DuffJ
 */

#define UMODE_INVISIBLE   0x00000001 /* makes user invisible */
#define UMODE_OPER        0x00000002 /* Operator */
#define UMODE_WALLOP      0x00000004 /* send wallops to them */
#define UMODE_FAILOP      0x00000008 /* Shows some global messages */
#define UMODE_REGNICK     0x00000020 /* Nick set by services as registered */
#define UMODE_SADMIN      0x00000040 /* Services Admin */
#define UMODE_ADMIN       0x00000080 /* Admin */

#define UMODE_SERVNOTICE  0x00000100 /* server notices such as kill */
#define UMODE_LOCOP       0x00000200 /* Local operator -- SRB */
#define UMODE_KILLS       0x00000400 /* Show server-kills... */
#define UMODE_FLOOD       0x00001000 /* Receive flood warnings */
#define UMODE_HMASK       0x00002000 /* mask part of hostname/IP */
#define UMODE_GONE        0x00004000 /* the user is away */
#define UMODE_UWNOTICE    0x00008000 /* receive WHOIS & USERHOST notices */

#define SEND_UMODES       (UMODE_INVISIBLE|UMODE_SERVNOTICE|UMODE_OPER|UMODE_WALLOP|UMODE_FAILOP|UMODE_REGNICK|UMODE_SADMIN|UMODE_ADMIN|UMODE_HMASK|UMODE_GONE|UMODE_UWNOTICE)
#define ALL_UMODES        (SEND_UMODES|UMODE_LOCOP|UMODE_KILLS|UMODE_FLOOD)
#define FLAGS_ID          (FLAGS_DOID|FLAGS_GOTID)

#define PROTO_NOQUIT      0x00000001 /* Negotiated NOQUIT protocol */
#define PROTO_TOKEN       0x00000002 /* Negotiated TOKEN protocol */
#define PROTO_ZIP         0x00000004 /* Negotiated ZIP protocol */

/* flags macros. */
#define IsKillsF(x)            ((x)->umodes & UMODE_KILLS)
#define IsFloodF(x)            ((x)->umodes & UMODE_FLOOD)
#define IsAdmin(x)             ((x)->umodes & UMODE_ADMIN)
#define IsSAdmin(x)            ((x)->umodes & UMODE_SADMIN)
#define SendFailops(x)         ((x)->umodes & UMODE_FAILOP)
#define IsOper(x)              ((x)->umodes & UMODE_OPER)
#define IsLocOp(x)             ((x)->umodes & UMODE_LOCOP)
#define IsInvisible(x)         ((x)->umodes & UMODE_INVISIBLE)
#define IsAnOper(x)            ((x)->umodes & (UMODE_OPER|UMODE_LOCOP))
#define IsARegNick(x)          ((x)->umodes & (UMODE_REGNICK))
#define IsRegNick(x)           ((x)->umodes & UMODE_REGNICK)
#define IsHMask(x)             ((x)->umodes & UMODE_HMASK)
#define IsPerson(x)            ((x)->user && IsClient(x))
#define IsPrivileged(x)        (IsAnOper(x) || IsServer(x))
#define SendWallops(x)         ((x)->umodes & UMODE_WALLOP)
#define SendServNotice(x)      ((x)->umodes & UMODE_SERVNOTICE && IsAnOper(x))
#define IsAway(x)              ((x)->umodes & UMODE_GONE)

#define IsListening(x)         ((x)->flags & FLAGS_LISTEN)
#define DoAccess(x)            ((x)->flags & FLAGS_CHKACCESS)
#define IsLocal(x)             ((x)->flags & FLAGS_LOCAL)
#define IsDead(x)              ((x)->flags & FLAGS_DEADSOCKET)
#define GotProtoctl(x)         ((x)->flags & FLAGS_PROTOCTL)
#define IsBlocked(x)           ((x)->flags & FLAGS_BLOCKED)

#define ZipIn(x)               ((x)->flags & FLAGS_ZIPPED_IN)
#define SetZipIn(x)            ((x)->flags |= FLAGS_ZIPPED_IN)

#define ZipOut(x)              ((x)->flags & FLAGS_ZIPPED_OUT)
#define SetZipOut(x)           ((x)->flags |= FLAGS_ZIPPED_OUT)

#define IsOutBurst(x)          ((x)->flags & FLAGS_OUT_BURST)
#define SetOutBurst(x)         ((x)->flags |= FLAGS_OUT_BURST)
#define ResOutBurst(x)         ((x)->flags &= ~FLAGS_OUT_BURST)

#define IsInBurst(x)           ((x)->flags & FLAGS_IN_BURST)
#define SetInBurst(x)          ((x)->flags |= FLAGS_IN_BURST)
#define ResInBurst(x)          ((x)->flags &= ~FLAGS_IN_BURST)

#define SetKillsF(x)           ((x)->umodes |= UMODE_KILLS)
#define SetFloodF(x)           ((x)->umodes |= UMODE_FLOOD)
#define SetOper(x)             ((x)->umodes |= UMODE_OPER)
#define SetLocOp(x)            ((x)->umodes |= UMODE_LOCOP)
#define SetInvisible(x)        ((x)->umodes |= UMODE_INVISIBLE)
#define SetWallops(x)          ((x)->umodes |= UMODE_WALLOP)
#define SetDNS(x)              ((x)->flags |= FLAGS_DOINGDNS)
#define DoingDNS(x)            ((x)->flags & FLAGS_DOINGDNS)
#define SetAccess(x)           ((x)->flags |= FLAGS_CHKACCESS)
#define SetBlocked(x)          ((x)->flags |= FLAGS_BLOCKED)
#define DoingAuth(x)           ((x)->flags & FLAGS_AUTH)
#define NoNewLine(x)           ((x)->flags & FLAGS_NONL)
#define SetRegNick(x)          ((x)->umodes & UMODE_REGNICK)

#define ClearAdmin(x)          ((x)->umodes &= ~UMODE_ADMIN)
#define ClearSAdmin(x)         ((x)->umodes &= ~UMODE_SADMIN)
#define ClearKillsF(x)         ((x)->umodes &= ~UMODE_KILLS)
#define ClearFloodF(x)         ((x)->umodes &= ~UMODE_FLOOD)
#define ClearFailops(x)        ((x)->umodes &= ~UMODE_FAILOP)
#define ClearOper(x)           ((x)->umodes &= ~UMODE_OPER)
#define ClearInvisible(x)      ((x)->umodes &= ~UMODE_INVISIBLE)
#define ClearWallops(x)        ((x)->umodes &= ~UMODE_WALLOP)
#define ClearDNS(x)            ((x)->flags &= ~FLAGS_DOINGDNS)
#define ClearAuth(x)           ((x)->flags &= ~FLAGS_AUTH)
#define ClearAccess(x)         ((x)->flags &= ~FLAGS_CHKACCESS)
#define ClearBlocked(x)        ((x)->flags &= ~FLAGS_BLOCKED)

/* ProtoCtl options */

#define DontSendQuit(x)        ((x)->proto & PROTO_NOQUIT)
#define SetNoQuit(x)           ((x)->proto |= PROTO_NOQUIT)
#define ClearNoQuit(x)         ((x)->proto &= ~PROTO_NOQUIT)

#define IsToken(x)             ((x)->proto & PROTO_TOKEN)
#define SetToken(x)            ((x)->proto |= PROTO_TOKEN)
#define ClearToken(x)          ((x)->proto &= ~PROTO_TOKEN)

/* works in conjunction with the Z flag in N: lines */
#define IsZipable(x)           ((x)->proto & PROTO_ZIP)
#define SetZipable(x)          ((x)->proto |= PROTO_ZIP)
#define ClearZipable(x)        ((x)->proto &= ~PROTO_ZIP)

#define OFLAG_OPER     0x00000001  /* IRC Operator */
#define OFLAG_LOCOP    0x00000002  /* Local Operator */
#define OFLAG_ADMIN    0x00000004  /* Admin */
#define OFLAG_SADMIN   0x00000008  /* Oper can be a services admin */

#define OPIsOper(x)            ((x)->oflag & OFLAG_OPER)
#define OPIsLocOp(x)           ((x)->oflag & OFLAG_LOCOP)
#define OPIsAnOper(x)          ((x)->oflag & (OFLAG_OPER|OFLAG_LOCOP))
#define OPIsAdmin(x)           ((x)->oflag & OFLAG_ADMIN)
#define OPIsSAdmin(x)          ((x)->oflag & OFLAG_SADMIN)

#define OPSSetOper(x)          ((x)->oflag |= OFLAG_OPER)
#define OPSSetLocOp(x)         ((x)->oflag |= OFLAG_LOCOP)
#define OPSSetAdmin(x)         ((x)->oflag |= OFLAG_ADMIN)
#define OPSSetSAdmin(x)        ((x)->oflag |= OFLAG_SADMIN)

#define OPClearOper(x)         ((x)->oflag &= ~OFLAG_OPER)
#define OPClearLocOp(x)        ((x)->oflag &= ~OFLAG_LOCOP)
#define OPClearAdmin(x)        ((x)->oflag &= ~OFLAG_ADMIN)
#define OPClearSAdmin(x)       ((x)->oflag &= ~OFLAG_SADMIN)

/*
 * defined debugging levels
 */
#define        DEBUG_FATAL  0
#define        DEBUG_ERROR  1  /* report_error() and other errors that are found */
#define        DEBUG_NOTICE 3
#define        DEBUG_DNS    4  /* used by all DNS related routines - a *lot* */
#define        DEBUG_INFO   5  /* general usful info */
#define        DEBUG_NUM    6  /* numerics */
#define        DEBUG_SEND   7  /* everything that is sent out */
#define        DEBUG_DEBUG  8  /* anything to do with debugging, ie unimportant :) */
#define        DEBUG_MALLOC 9  /* malloc/free calls */
#define        DEBUG_LIST  10  /* debug list use */

/*
 * defines for curses in client
 */
#define        DUMMY_TERM      0
#define        CURSES_TERM     1
#define        TERMCAP_TERM    2

struct  SqlineItem     {
	u_int32_t status;
	char *sqline;
	char *reason;
	struct  SqlineItem *next;
};

struct ConfItem {
	u_int32_t status; /* If CONF_ILLEGAL, delete when no clients */
	u_int32_t flags; /* Additional flags */
	int clients; /* Number of *LOCAL* clients using this */
	struct in_addr ipnum; /* ip number of host field */
	char *host;
	char *passwd;
	char *name;
	int port;
	time_t hold;   /* Hold action until this time (calendar time) */
	int tmpconf;
#ifndef VMSP
	aClass *class;  /* Class of connection */
#endif
	struct ConfItem *next;
};

#define        CONF_ILLEGAL            0x80000000
#define        CONF_MATCH              0x40000000
#define        CONF_QUARANTINED_SERVER 0x0001
#define        CONF_CLIENT             0x0002
#define        CONF_CONNECT_SERVER     0x0004
#define        CONF_NOCONNECT_SERVER   0x0008
#define        CONF_LOCOP              0x0010
#define        CONF_OPERATOR           0x0020
#define        CONF_ME                 0x0040
#define        CONF_KILL               0x0080
#define        CONF_ADMIN              0x0100
#ifdef         R_LINES
#define        CONF_RESTRICT           0x0200
#endif
#define        CONF_CLASS              0x0400
#define        CONF_SERVICE            0x0800
#define        CONF_LEAF               0x1000
#define        CONF_LISTEN_PORT        0x2000
#define        CONF_HUB                0x4000
#define        CONF_UWORLD             0x8000
#define        CONF_QUARANTINED_NICK  0x10000
#define        CONF_ZAP               0x20000
#define        CONF_CONFIG             0x100000
#define        CONF_CRULEALL           0x200000
#define        CONF_CRULEAUTO          0x400000
#define        CONF_MISSING           0x800000
#define        CONF_SADMIN            0x1000000
#define        CONF_DRPASS            0x2000000   /* DIE/RESTART pass - NikB */
/* Borg88
Extra CONF lines
*/
#define        CONF_VHOST             0x4000000

#define        CONF_OPS                (CONF_OPERATOR | CONF_LOCOP)
#define        CONF_SERVER_MASK        (CONF_CONNECT_SERVER | CONF_NOCONNECT_SERVER)
#define        CONF_CLIENT_MASK        (CONF_CLIENT | CONF_SERVICE | CONF_OPS | CONF_SERVER_MASK)
#define        CONF_CRULE              (CONF_CRULEALL | CONF_CRULEAUTO)
#define        CONF_QUARANTINE         (CONF_QUARANTINED_SERVER|CONF_QUARANTINED_NICK)

#define        IsIllegal(x)    ((x)->status & CONF_ILLEGAL)
#define IsTemp(x)      ((x)->tmpconf)

/* Borg88
Extra flags for ConfItem
*/
#define FLAG_TRIED 1

/* support for unicode */
typedef struct aCharSetAlias CharSetAlias;
typedef struct aCharSet CharSet;

struct aCharSetAlias {
	struct aCharSetAlias *next;
	char *name;
};

struct aCharSet {
	/* charset name and aliases */
	char *name;
	CharSetAlias *alias;

	/* for maintenance purposes */
	struct aCharSet *next, *prev;

	unsigned int refcount;

	/* unicode boundaries */
	unsigned short ucmin, ucmax;
	/* ascii-to-unicode map */
	unsigned short a2u[256];
	/* unicode-to-ascii map */
	unsigned char u2a[1];
};

typedef struct {
	/* boundaries */
	unsigned short ucmin, ucmax;

	/* sequence map */
	char *map, *uc2seq[1];
} TransLitSet;


/* Client structures */
struct User {
       struct  User *nextu;
       Link    *channel;       /* chain of channel pointer blocks */
       Link    *invited;       /* chain of invite pointer blocks */
       Link    *silence;       /* chain of silence pointer blocks */
       time_t  last;
       u_int32_t servicestamp;  /* Services' time stamp variable */
       int     refcnt;         /* Number of times this block is referenced */
       int     joined;         /* number of channels joined */
       char    username[USERLEN+1];
       char    host[HOSTLEN+1];
	   char    vhost[HOSTLEN+1]; /* masked host, +x umode */
       /* char   server[HOSTLEN+1]; */
                               /*
                               ** In a perfect world the 'server' name
                               ** should not be needed, a pointer to the
                               ** client describing the server is enough.
                               ** Unfortunately, in reality, server may
                               ** not yet be in links while USER is
                               ** introduced... --msa
                               ** No.  We are going to assume the server
                               ** is there.  There is no good reason at
                               ** all that we should get a NICK line
                               ** before a SERVER line.   -- Aeto
                               ** BUT, it is hiddenin aClient! -- Aeto
                               */
#ifdef LIST_DEBUG
       aClient *bcptr;
#endif
};

struct Server  {
	struct Server  *nexts;
	struct User *user;          /* who activated this connection */
	char   up[HOSTLEN+1];  /* uplink for this server */
	char   by[NICKLEN+1];
	struct ConfItem *nline;       /* N-line pointer for this server */
#ifdef LIST_DEBUG
	struct Client *bcptr;
#endif
	void   *zip_out;
	void   *zip_in;
};

struct Client  {
       struct  Client *next, *prev, *hnext;
       anUser  *user;          /* ...defined, if this is a User */
       aServer *serv;          /* ...defined, if this is a server */
       time_t  lasttime;       /* ...should be only LOCAL clients? --msa */
       time_t  firsttime;      /* time client was created */
       time_t  since;          /* last time we parsed something */
       time_t  lastnick;       /* TimeStamp on nick */
       time_t  nextnick;       /* Time the next nick change will be allowed */
       time_t  nexttarget;     /* Time until a change in targets is allowed */
       u_char targets[MAXTARGETS]; /* Hash values of current targets */
       long    flags;          /* client flags */
       long    umodes;         /* client usermodes */
       aClient *from;          /* == self, if Local Client, *NEVER* NULL! */
       int     fd;             /* >= 0, for local clients */
       int     hopcount;       /* number of servers to this 0 = local */
       short   status;         /* Client type */
       char    name[HOSTLEN+1];/* Unique name of the client, nick or host */
       char    username[USERLEN+1]; /* username here now for auth stuff */
       char    info[REALLEN+1];/* Free form additional client information */
       struct  Client *srvptr; /* Server introducing this.  May be &me */
       Link    *history;       /* Whowas linked list */
       int     whox_mark;      /* for WHO processing */
       /*
       ** The following fields are allocated only for local clients
       ** (directly connected to *this* server with a socket.
       ** The first of them *MUST* be the "count"--it is the field
       ** to which the allocation is tied to! *Never* refer to
       ** these fields, if (from != self).
       */
       int     count;          /* Amount of data in buffer */
       /* Incoming message buffer
       char    buffer[BUFSIZE];
       */
       char    buffer[MAXMSGLEN];/* extended to fit 512 utf chars, not bytes */
       u_short lastsq;         /* # of 2k blocks when sendqueued called last*/
       dbuf    sendQ;          /* Outgoing message queue--if socket full */
       dbuf    recvQ;          /* Hold for data incoming yet to be parsed */
       long    oflag;          /* Operator access flags -Cabal95 */
       long    proto;          /* ProtoCtl options */
       long    sendM;          /* Statistics: protocol messages send */
       long    sendK;          /* Statistics: total k-bytes send */
       long    receiveM;       /* Statistics: protocol messages received */
       long    receiveK;       /* Statistics: total k-bytes received */
       u_short sendB;          /* counters to count upto 1-k lots of bytes */
       u_short receiveB;       /* sent and received. */
       aClient *acpt;          /* listening client which we accepted from */
       Link    *confs;         /* Configuration record associated */
       int     authfd;         /* fd for rfc931 authentication */
       struct  in_addr ip;     /* keep real ip# too */
       u_short port;           /* and the remote port# too :-) */
       struct  hostent *hostp;
       u_short notifies;       /* Keep track of count of notifies */
       Link    *notify;        /* Links to clients notify-structures */
       LOpts   *lopt;          /* Saved /list options */
#ifdef pyr
       struct  timeval lw;
#endif
       char    *away;          /* pointer to away message */
       char    sockhost[HOSTLEN+1]; /* This is the host name from the socket
                                    ** and after which the connection was
                                    ** accepted.
                                    */
       char    vhost[HOSTLEN+1];
       char    passwd[PASSWDLEN+1];
#ifdef DEBUGMODE
       time_t  cputime;
#endif
	CharSet *charset;
};

#define        CLIENT_LOCAL_SIZE sizeof(aClient)
#define        CLIENT_REMOTE_SIZE offsetof(aClient,count)

/*
 * statistics structures
 */
struct stats {
       unsigned int    is_cl;  /* number of client connections */
       unsigned int    is_sv;  /* number of server connections */
       unsigned int    is_ni;  /* connection but no idea who it was */
       unsigned short  is_cbs; /* bytes sent to clients */
       unsigned short  is_cbr; /* bytes received to clients */
       unsigned short  is_sbs; /* bytes sent to servers */
       unsigned short  is_sbr; /* bytes received to servers */
       unsigned long   is_cks; /* k-bytes sent to clients */
       unsigned long   is_ckr; /* k-bytes received to clients */
       unsigned long   is_sks; /* k-bytes sent to servers */
       unsigned long   is_skr; /* k-bytes received to servers */
       time_t          is_cti; /* time spent connected by clients */
       time_t          is_sti; /* time spent connected by servers */
       unsigned int    is_ac;  /* connections accepted */
       unsigned int    is_ref; /* accepts refused */
       unsigned int    is_unco; /* unknown commands */
       unsigned int    is_wrdi; /* command going in wrong direction */
       unsigned int    is_unpf; /* unknown prefix */
       unsigned int    is_empt; /* empty message */
       unsigned int    is_num; /* numeric message */
       unsigned int    is_kill; /* number of kills generated on collisions */
       unsigned int    is_fake; /* MODE 'fakes' */
       unsigned int    is_asuc; /* successful auth requests */
       unsigned int    is_abad; /* bad auth requests */
       unsigned int    is_udp; /* packets recv'd on udp port */
       unsigned int    is_loc; /* local connections made */
};

struct ListOptions {
       LOpts   *next;
       Link    *yeslist, *nolist;
       int     starthash;
       short int       showall;
       unsigned short  usermin;
       int     usermax;
       time_t  currenttime;
       time_t  chantimemin;
       time_t  chantimemax;
       time_t  topictimemin;
       time_t  topictimemax;
};

/* mode structure for channels */

struct SMode {
	unsigned int mode;
	int limit;
	char key[KEYLEN+1];
};

/* Message table structure */

struct Message {
       char    *cmd;
       int     (* func)();
       unsigned int    count;
       int     parameters;
       char    flags;
               /* bit 0 set means that this command is allowed to be used
                * only on the average of once per 2 seconds -SRB */
       u_char  token[2]; /* Cheat for tokenized value */
       unsigned long   bytes;
#ifdef DEBUGMODE
       unsigned long   lticks;
       unsigned long   rticks;
#endif
};

/* Used for notify-hash buckets... -Donwulff */

struct Notify {
       aNotify *hnext;
       time_t  lasttime;
       Link    *notify;
       char    nick[1];
};

/* general link structure used for chains */

struct SLink   {
	struct SLink *next;
	u_int32_t flags;
	union {
		aClient *cptr;
		aChannel *chptr;
		aConfItem *aconf;
		aNotify *nptr;
		aName *whowas;
		char    *cp;
	} value;
};

struct SBan    {
       struct  SBan    *next;
       char *banstr;
       char *who;
       time_t when;
};

/* channel structure */

struct Channel {
       struct  Channel *nextch, *prevch, *hnextch;
       Mode    mode;
       time_t  creationtime;
       char    topic[TOPICLEN+1];
       char    topic_nick[NICKLEN+1];
       time_t  topic_time;
       int     users;
       Link    *members;
       Link    *invites;
       Ban     *banlist;
       Ban     *exclist;
       Ban     *invlist;
       char    chname[1];
};

/*
** Channel Related macros follow
*/

/* Channel related flags */

#define        CHFL_CHANOP     0x0001 /* Channel operator */
#define        CHFL_VOICE      0x0002 /* the power to speak */
#define        CHFL_DEOPPED    0x0004 /* Is de-opped by a server */
#define        CHFL_SERVOPOK   0x0008 /* Server op allowed */
#define        CHFL_ZOMBIE     0x0010 /* Kicked from channel */
/* Bans are stored in separate linked list, so phase this out? */
#define        CHFL_BAN        0x0020 /* ban channel flag */
#define        CHFL_OVERLAP    (CHFL_CHANOP|CHFL_VOICE)

/* Channel Visibility macros */

#define        MODE_CHANOP     CHFL_CHANOP
#define        MODE_VOICE      CHFL_VOICE
#define        MODE_PRIVATE    0x00000004
#define        MODE_SECRET     0x00000008
#define        MODE_MODERATED  0x00000010
#define        MODE_TOPICLIMIT 0x00000020
#define        MODE_INVITEONLY 0x00000040
#define        MODE_NOPRIVMSGS 0x00000080
#define        MODE_KEY        0x00000100
#define        MODE_BAN        0x00000200
#define        MODE_LIMIT      0x00000400
#define        MODE_RGSTR      0x00000800
#define        MODE_RGSTRONLY  0x00001000
#define        MODE_EXC        0x00002000
#define        MODE_INV        0x00004000
#define        MODE_NOCOLOR    0x00008000
#define        MODE_JUPED      0x00010000
#define        MODE_FORWARD    0x00020000

/* mode flags which take another parameter (With PARAmeterS) */
#define        MODE_WPARAS     (MODE_CHANOP|MODE_VOICE|MODE_BAN|MODE_EXC|MODE_INV|MODE_KEY|MODE_LIMIT|MODE_FORWARD)
/*
 * Undefined here, these are used in conjunction with the above modes in
 * the source.
#define        MODE_DEL       0x40000000
#define        MODE_ADD       0x80000000
 */

#define HoldChannel(x)          (!(x))

/* hidden or private channel */
#define SecretChannel(x) ((x) && ((x)->mode.mode & (MODE_PRIVATE|MODE_SECRET)))

/* disabled channel */
#define JupedChannel(x) ((x) && ((x)->mode.mode & MODE_JUPED))

/* colour-blind channels */
#define ColorBlindChannel(x)   ((x) && ((x)->mode.mode & MODE_NOCOLOR))

/* persistant channel - is not removed after last user parts */
#define PersistantChannel(x) ((x)->mode.mode & (MODE_RGSTR|MODE_JUPED|MODE_FORWARD))

/* channel visible */
#define ShowChannel(v,c) (!SecretChannel(c) || IsMember((v),(c)) || IsAnOper(v))

#define IsChannelName(name) ((name) && (*(name) == '#' || *(name) == '&' || *(name) == '+'))
#define IsModelessChannel(name) ((name) && (*(name) == '+'))

/* Misc macros */

#define        BadPtr(x) (!(x) || (*(x) == '\0'))

#define        isvalid(c) (((c) >= 'A' && (c) <= '~') || isdigit(c) || (c) == '-')

#define        MyConnect(x)                    ((x)->fd >= 0)
#define        MyClient(x)                     (MyConnect(x) && IsClient(x))
#define        MyOper(x)                       (MyConnect(x) && IsOper(x))

/* Lifted somewhat from Undernet code --Rak */

#define IsSendable(x)          (DBufLength(&x->sendQ) < 2048)
#define DoList(x)              ((x)->lopt)

/* String manipulation macros */

#define strncpyzt strlcpy
/* #define strncpyzt(x, y, N) do{(void)strncpy(x,y,N);*(x+N-1)='\0';}while(0) */
#define StrEq(x,y) (!strcmp((x),(y)))

/* used in SetMode() in channel.c and m_umode() in s_msg.c */

#define        MODE_NULL      0
#define        MODE_ADD       0x40000000
#define        MODE_DEL       0x20000000

/* return values for hunt_server() */

#define        HUNTED_NOSUCH   (-1)    /* if the hunted server is not found */
#define        HUNTED_ISME     0       /* if this server should execute the command */
#define        HUNTED_PASS     1       /* if message passed onwards successfully */

/* used when sending to #mask or $mask */

#define        MATCH_SERVER  1
#define        MATCH_HOST    2

/* used for async dns values */

#define        ASYNC_NONE      (-1)
#define        ASYNC_CLIENT    0
#define        ASYNC_CONNECT   1
#define        ASYNC_CONF      2
#define        ASYNC_SERVER    3

/* misc variable externs */

extern const char *version, *creation;

/* misc defines */

#define ZIP_NEXT_BUFFER -4
#define FLUSH_BUFFER    -2
#define UTMP            "/etc/utmp"
#define COMMA           ","

#define IRCERR_ZIP -12

#endif /* __struct_include__ */
