/* m_codepage.c -- change current codepage
 * Copyright (C) 2001-2002, ForestNet IRC Network
 * This is a free software; see file `license' for details.
 * $Id: m_codepage.c,v 1.21 2002/11/18 12:44:07 hex Exp $
 */

#include "struct.h"
#include "send.h"
#include "numeric.h"
#include "h.h"
#include "msg.h"
#include "channel.h"
#include "unicode.h"

static char *cp_unicode[] = { "utf8", "utf-8", "unicode", "none", NULL };

/* parv[0] - codepage name, parv[1] - nick to set codepage for */
int m_codepage(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	char **p, tmpbuf[512];
	aClient *acptr;
	CharSet *cs;

	if (check_registered(sptr))
		return 0;

	if (parc < 2) {
		send_syntax(sptr, MSG_CHARSET, "CHARSET name [nick] "
			"(current: %s)", (sptr->charset == NULL)
			? "utf-8" : sptr->charset->name);
		return 0;
	}

	if (parc > 2 && MyClient(sptr) && !IsOper(sptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	if (parc > 2 && hunt_server(cptr, sptr, ":%s CODEPAGE %s %s",
		2, parc, parv) != HUNTED_ISME) return 0;

	/* hard-coded alias for backwards compatibility */
	if (strcasecmp(parv[1], "translit") == 0)
		parv[1] = "us_ascii";

	acptr = (parc > 2) ? find_person(parv[2], 0) : sptr;

	for (p = cp_unicode; *p != NULL; ++p) {
		if (strcasecmp(parv[1], *p) == 0) {
			/* if charset is not changed - don't send notice */
			if (acptr->charset == NULL)
				return 0;

			cs = NULL;
			break;
		}
	}

	/* it was not a request for unicode charset, let's find something. */
	if ((*p == NULL) && (cs = charset_attach(parv[1])) == NULL) {
		sendto_one(sptr, ":%s NOTICE %s :Unknown charset \"%s\".",
			me.name, sptr->name, parv[1]);
		return 0;
	}

	if (acptr->charset != NULL)
		charset_detach(acptr->charset);

	acptr->charset = cs;

	if (sptr == acptr)
		tmpbuf[0] = '\0';
	else {
		strcpy(tmpbuf, " by ");
		strcat(tmpbuf, sptr->name);
	}

	sendto_one(acptr, ":%s NOTICE %s :Your charset was changed "
		"to %s%s.", me.name, acptr->name, (cs == NULL)
		? "utf-8" : cs->name, tmpbuf);

	return 0;
}
