/* m_kill.c -- remove a klient from the network;
 * Copyright (C) 2001-2002, ForestNet IRC Network
 * This is a free software; see file `license' for details.
 * $Id: m_kill.c,v 1.2 2002/11/18 12:44:08 hex Exp $
 */

#include <stdlib.h>
#include <string.h>

#include "struct.h"
#include "common.h"
#include "numeric.h"
#include "send.h"
#include "h.h"
#include "ircdlog.h"


static anUser UnknownUser = {
	NULL,			/* nextu */
	NULL,			/* channel */
	NULL,			/* invited */
	NULL,			/* silence */
	0,			/* last */
	0,			/* servicestamp */
	1,			/* refcount */
	0,			/* joined */
	"<Unknown>",		/* username */
	"<Unknown>",		/* host */
	"<Unknown>"		/* server */
};


/* m_kill
 *     parv[0] = sender prefix
 *     parv[1] = kill victim(s) - comma separated list
 *     parv[2] = kill path
 */
int m_kill(aClient *cptr, aClient * sptr, int parc, char *parv[])
{
	anUser *auser;
	aClient *acptr;
	int chasing = 0, kcount = 0;
	char inpath[HOSTLEN * 2 + USERLEN + 5], *user,
		*path, *killer, *nick, *p, *s;
	char buf[1024], tmpbuf[1024];

	if (check_registered(sptr))
		return 0;

	if (parc < 2 || *parv[1] == '\0') {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
			me.name, parv[0], "KILL");
		return 0;
	}

	if (!IsPrivileged(cptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	user = parv[1];
	path = parv[2]; /* Either defined or NULL (parc >= 2!!) */

	strcpy(inpath, get_client_name(cptr, FALSE));

	/* truncate the server name at the first dot.  this makes paths look
	 * better for networks with good server naming conventions, but at
	 * networks like IRCnet it's undesirable.
	 */
	if (IsServer(cptr) && (s = strchr(inpath, '.')) != NULL)
		*s = '\0';

	if (IsAnOper(cptr)) {
		if (BadPtr(path)) {
			sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
				me.name, parv[0], "KILL");
			return 0;
		}
		if (strlen(path) > (size_t) TOPICLEN)
			path[TOPICLEN] = '\0';
	}

	if (MyClient(sptr))
		user = canonize(user);

	nick = strtoken(&p, user, ",");

	for (p = NULL; nick; nick = strtoken(&p, NULL, ",")) {
		chasing = 0;

		if (!(acptr = find_client(nick, NULL))) {
			/* If the user has recently changed nick, we
			 * automaticly rewrite the KILL for this new
			 * nickname -- this keeps servers in synch when
			 * nick change and kill collide.
			 */
			if (!(acptr = get_history(nick,
			(long)KILLCHASETIMELIMIT))) {
				sendto_one(sptr, err_str(ERR_NOSUCHNICK),
					me.name, parv[0], nick);
				continue;
			}
			sendto_one(sptr, ":%s NOTICE %s :KILL changed from "
				"%s to %s", me.name, parv[0],
				nick, acptr->name);
			chasing = 1;
		}

		if (!MyConnect(acptr) && MyClient(cptr) && !OPIsOper(cptr)) {
			sendto_one(sptr, err_str(ERR_NOPRIVILEGES),
				me.name, parv[0]);
			continue;
		}

		if (IsServer(acptr) || IsMe(acptr)) {
			sendto_one(sptr, err_str(ERR_CANTKILLSERVER),
				me.name, parv[0]);
			continue;
		}

		/* From here on, the kill is probably going to be successful. */

		++kcount;

		if (!IsServer(sptr) && (kcount > MAXKILLS)) {
			sendto_one(sptr, ":%s NOTICE %s :Too many targets, "
				"kill list was truncated. Maximum is %d.",
				me.name, parv[0], MAXKILLS);
			break;
		}

		if (!IsServer(cptr)) {
			/* The kill originates from this server, initialize
			 * path. (In which case the 'path' may contain user
			 * suplied explanation ...or some nasty comment,
			 * sigh... >;-)
			 * ...!operhost!oper ...!operhost!oper (comment)
			 */
			strcpy(inpath, cptr->sockhost);

			/* Only check the path the first time around,
			 * or it gets appended to itself.
			 */
			if (kcount < 2)
				if (!BadPtr(path)) {
					sprintf(buf, "%s%s (%s)", cptr->name,
						IsOper(sptr) ? "" : "(L)",
						path);
					path = buf;
				}
				else
					path = cptr->name;
		}
		else if (BadPtr(path))
			path = "*no-path*";

		auser = acptr->user ? acptr->user : &UnknownUser;

		snprintf(tmpbuf, sizeof(tmpbuf),
			"%s!%s@%s from %s, path: %s!%s",
			acptr->name, auser->username, auser->host,
			parv[0], inpath, path);

		sendto_umode(UMODE_KILLS, "*** Notice -- Received KILL "
			"message for %s", tmpbuf);

		sendto_syschan("kills", "%s", tmpbuf);

		putlog(LOG_INFO, "KILL: %s", tmpbuf);

		/* And pass on the message to other servers.  Note, that if
		 * KILL was changed, the message has to be sent to all links,
		 * also back.  Suicide kills are NOT passed on --SRB
		 */
		if (!MyConnect(acptr) || !MyConnect(sptr) || !IsAnOper(sptr)) {
			sendto_serv_butone(cptr, ":%s KILL %s :%s!%s",
				parv[0], acptr->name, inpath, path);
			if (chasing && IsServer(cptr))
				sendto_one(cptr, ":%s KILL %s :%s!%s",
					me.name, acptr->name, inpath, path);
			acptr->flags |= FLAGS_KILLED;
		}

		if (MyConnect(acptr) && MyConnect(sptr) && IsAnOper(sptr))
			sprintf(tmpbuf, "Local kill by %s (%s)", sptr->name,
				BadPtr(parv[2]) ? sptr->name : parv[2]);
		else {
			if ((killer = strchr(path, ' '))) {
				while (*killer != '\0' && *killer != '!')
					--killer;
				if (!*killer)
					killer = path;
				else
					++killer;
			}
			else
				killer = path;

			sprintf(tmpbuf, "Killed (%s)", killer);
		}

		if (exit_client(cptr, acptr, sptr, tmpbuf) == FLUSH_BUFFER)
			return FLUSH_BUFFER;
	}

	return 0;
}
