/* m_oper.c -- get IRC Operator priviledes;
 * Copyright (C) 2001-2002, ForestNet IRC Network
 * This is a free software; see file `license' for details.
 * $Id: m_oper.c,v 1.20 2002/11/18 12:44:08 hex Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "struct.h"
#include "send.h"
#include "numeric.h"
#include "h.h"
#include "ircdmd5.h"
#include "common.h" /* for opers_log */
#include "ircdlog.h"

static void log_attempt(const char *uid, aClient *sptr, int success)
{
	FILE *log;

	if (opers_log == NULL)
		return;

	if (!strcmp(opers_log, "syslog")) {
		putlog(success ? LOG_NOTICE : LOG_WARNING, "OPER %s attempt "
			"by %s, %s.", uid, get_client_name(sptr, TRUE),
			success ? "succeeded" : "FAILED");
		return;
	}

	if ((log = fopen(opers_log, "a")) != NULL) {
		fprintf(log, "%s OPER %s attempt by %s, %s.\n",
			my_timestamp(), uid, get_client_name(sptr, TRUE),
			success ? "succeeded" : "FAILED");
		fclose(log);
	}
}


static int check_oper_password(aClient *cptr, aConfItem *aconf, char *passwd)
{
	if (!(aconf->status & CONF_OPS))
		return 0;

	if (strncmp(aconf->passwd, "$1$", 3))
		return 0;

	if (strcmp(md5crypt(passwd, aconf->passwd), aconf->passwd) != 0)
		return 0;

	if (attach_conf(cptr, aconf) != 0)
		return 0;

	return 1;
}


/* parv[0] = sender prefix, parv[1] = oper name, parv[2] = oper password */
int m_oper(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aConfItem *aconf;

	if (check_registered_user(sptr))
		return 0;

	if (MyConnect(sptr) && IsAnOper(sptr)) {
		sendto_one(sptr, ":%s 381 :You are an IRC Operator already!",
			me.name, parv[0]);
		return 0;
	}

	if (!IsServer(cptr) && (BadPtr(parv[1]) || BadPtr(parv[2])
	|| parc < 3)) {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
			me.name, parv[0], "OPER");
		return 0;
	}

	/* if message arrived from server, trust it, and set to oper */
	if ((IsServer(cptr) || IsMe(cptr)) && !IsOper(sptr)) {
		sptr->umodes |= UMODE_OPER;
		sendto_serv_butone(cptr, ":%s MODE %s :+o", parv[0], parv[0]);
		if (IsMe(cptr))
			sendto_one(sptr, rpl_str(RPL_YOUREOPER),
				me.name, parv[0]);
		return 0;
	}

	aconf = find_conf_exact(parv[1], sptr->username,
		(*sptr->vhost) ? sptr->vhost : sptr->sockhost, CONF_OPS);

	if (aconf == NULL)
		aconf = find_conf_exact(parv[1], sptr->username,
			inetntoa((char*)&cptr->ip), CONF_OPS);

	if (aconf == NULL) {
		sendto_one(sptr, err_str(ERR_NOOPERHOST), me.name, parv[0]);
		sendto_realops("Failed OPER attempt by %s[%s@%s]",
			parv[0], sptr->user->username, (*sptr->vhost)
			? sptr->vhost : sptr->sockhost);
		sptr->since += 7;
		log_attempt(parv[1], sptr, FALSE);
		return 0;
	}

	if (check_oper_password(sptr, aconf, parv[2])) {
		int old = (sptr->umodes & ALL_UMODES);
		char *s;

		if (s = strchr(aconf->host, '@'))
			*s++ = '\0';

		if (!(aconf->port & OFLAG_OPER))
			SetLocOp(sptr);
		else
			SetOper(sptr);

		sptr->oflag = aconf->port;
		*--s = '@';
		sendto_ops("%s (%s!%s@%s) is now operator (%c)", parv[1],
			parv[0], sptr->user->username, sptr->user->host,
			IsOper(sptr) ? 'o' : 'O');
		sptr->umodes |= (UMODE_SERVNOTICE|UMODE_WALLOP
			|UMODE_FAILOP|UMODE_FLOOD);
		if (sptr->oflag & OFLAG_ADMIN)
			sptr->umodes |= UMODE_ADMIN;
		if (sptr->oflag & OFLAG_SADMIN)
			sptr->umodes |= UMODE_SADMIN;
		send_umode_out(cptr, sptr, old);
		sendto_one(sptr, rpl_str(RPL_YOUREOPER), me.name, parv[0]);

		log_attempt(parv[1], sptr, TRUE);
	}
	else {
		detach_conf(sptr, aconf);
		sendto_one(sptr,err_str(ERR_PASSWDMISMATCH),me.name, parv[0]);

		sendto_realops("Failed OPER attempt by %s[%s@%s] using UID %s",
			parv[0], sptr->user->username,
			(*sptr->vhost) ? sptr->vhost : sptr->sockhost, parv[1]);

		sendto_serv_butone(&me, ":%s GLOBOPS :Failed OPER attempt by "
			"%s[%s@%s] using UID %s", me.name, parv[0],
			sptr->user->username,
			(*sptr->vhost) ? sptr->vhost : sptr->sockhost, parv[1]);

		sptr->since += 7;

		log_attempt(parv[1], sptr, FALSE);
	}

	hmask_update(sptr);

	return 0;
}
