/* m_services.c -- various functions to support services;
 * Copyright (C) 2001-2002, ForestNet IRC Network
 * This is a free software; see file `license' for details.
 * $Id: m_services.c,v 1.10 2002/11/18 12:44:08 hex Exp $
 */

#include "config.h"

#include "struct.h"
#include "send.h"
#include "numeric.h"
#include "h.h"

/* Built in services aliasing for ChanServ, Memoserv, NickServ and
 * OperServ. This not only is an alias, but is also a security measure,
 * because PRIVMSG's arent sent to 'ChanServ' they are now sent to
 * 'ChanServ@services.dal.net' so nobody can snoop /cs commands :) -taz
 */

int m_chanserv(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	if (check_registered_user(sptr))
		return 0;

	if (parc < 2 || BadPtr(parv[1])) {
		sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
		return -1;
	}

	if (BadPtr(services_host) || !csptr)
		sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name,
			parv[0], cs_nick);
	else if (csptr)
		sendto_one(csptr, ":%s PRIVMSG %s@%s :%s", parv[0],
			cs_nick, services_host, parv[1]);

	return 0;
}


int m_memoserv(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	if (check_registered_user(sptr))
		return 0;

	if (parc < 2 || BadPtr(parv[1])) {
		sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
		return -1;
	}

	if (BadPtr(services_host) || !msptr)
		sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name,
			parv[0], ms_nick);
	else if (msptr)
		sendto_one(msptr, ":%s PRIVMSG %s@%s :%s", parv[0],
			ms_nick, services_host, parv[1]);

	return 0;
}


int m_nickserv(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	if (check_registered_user(sptr))
		return 0;

	if (parc < 2 || BadPtr(parv[1])) {
		sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
		return -1;
	}

	if (BadPtr(services_host) || !nsptr)
		sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name,
			parv[0], ns_nick);
	else if (nsptr)
		sendto_one(nsptr, ":%s PRIVMSG %s@%s :%s", parv[0],
			ns_nick, services_host, parv[1]);

	return 0;
}


int m_operserv(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	if (check_registered_user(sptr))
		return 0;

	if (parc < 2 || BadPtr(parv[1])) {
		sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
		return -1;
	}

	if (BadPtr(services_host) || !osptr)
		sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name,
			parv[0], os_nick);
	else if (osptr)
		sendto_one(osptr, ":%s PRIVMSG %s@%s :%s", parv[0],
			os_nick, services_host, parv[1]);

	return 0;
}


/* Automatic NickServ/ChanServ parsing. If the second word of parv[1]
 * starts with a '#' this message goes to ChanServ. If it starts with
 * anything else, it goes to NickServ. If there is no second word in
 * parv[1], the message defaultly goes to NickServ. If parv[1] == 'help'
 * the user in instructed to /cs, /ns or /ms HELP for the help they need.
 * -taz
 */

int m_services(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	char *tmps;

	if (check_registered_user(sptr))
		return 0;

	if (parc < 2 || BadPtr(parv[1])) {
		sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
		return -1;
	}

	if ((strlen(parv[1]) >= 4) && (!strncmp(parv[1], "help", 4))) {
		sendto_one(sptr, ":services!service@%s NOTICE %s :For ChanServ "
			"help use: /quote cs help", services_host, sptr->name);
		sendto_one(sptr, ":services!service@%s NOTICE %s :For NickServ "
			"help use: /quote ns help", services_host, sptr->name);
		sendto_one(sptr, ":services!service@%s NOTICE %s :For MemoServ "
			"help use: /quote ms help", services_host, sptr->name);
		return 0;
	}

	if ((tmps = (char *)strchr(parv[1], ' '))) {
		tmps++;
		if (*tmps == '#')
			return m_chanserv(cptr, sptr, parc, parv);
		else
			return m_nickserv(cptr, sptr, parc, parv);
	}

	return m_nickserv(cptr, sptr, parc, parv);
}
