/* m_who.c -- list clients matching a specific criteria;
 * Copyright (C) 2001-2002, ForestNet IRC Network
 * This is a free software; see file `license' for details.
 * $Id: m_who.c,v 1.19 2002/11/18 12:44:08 hex Exp $
 */

#include <string.h>

#include "config.h"
#include "struct.h"
#include "send.h"
#include "numeric.h"
#include "h.h"
#include "channel.h"
#include "common.h"

static int whox_mark, show_real_host;

static void do_who(aClient *sptr, aClient *acptr, aChannel *repchan, char *mask)
{
	int i = 0, show, su;
	char status[5];

	/* check if the sender has extra privilege over this client */
	su = (sptr == acptr || IsOper(sptr)
		|| (IsLocOp(sptr) && MyClient(acptr))) ? 1 : 0;

	/* mark this client as already processed */
	acptr->whox_mark = whox_mark;

	if (mask && *mask) {
		char host[2][NICKLEN + USERLEN + HOSTLEN + 6];

		make_nick_user_hosts(acptr, host[0], host[1]);

		/* hide unwanted clients */
		if (*mask == '!') {
			++mask;

			/* an oper wishing not to see clients on
			 * a specific server */
			if (su && !match(mask, acptr->srvptr->name))
				return;

			/* don't want to see clients matching this */
			if (!match(mask, IsHMask(acptr) ? host[1] : host[0]))
				return;

			/* for opers, also check the real host */
			if (IsHMask(acptr) && su && !match(mask, host[0]))
				return;
		}

		/* reversed match */
		else {
			show = 0;

			/* I want this server! */
			if (su && !match(mask, acptr->srvptr->name))
				show = 1;

			/* I want this client! */
			if (!show && !match(mask, IsHMask(acptr)
			? host[1] : host[0]))
				show = 1;

			/* I'm an oper and I want this _real_ host */
			if (!show && su && IsHMask(acptr)
			&& !match(mask, host[0]))
				show = 1;

			if (!show)
				return;
		}
	}

	/* if we're not listing a specific channel (and thus have already
	 * checked that the sender is allowed to see the client), make sure
	 * they share some channel and can be shown */
	if (!su && !repchan && IsInvisible(acptr)) {
		Link *lp;

		for (show = 0, lp = acptr->user->channel; lp; lp = lp->next)
			if (IsMember(sptr, lp->value.chptr)) {
				show = 1;
				break;
			}

		if (!show)
			return;
	}

	status[i++] = IsAway(acptr) ? 'G' : 'H';

	if (IsAnOper(acptr))
		status[i++] = '*';
	else if (IsInvisible(acptr) && sptr != acptr && IsAnOper(sptr))
		status[i++] = '%';

	if (repchan && has_op(acptr, repchan))
		status[i++] = '@';
	else if (repchan && has_voice(acptr, repchan))
		status[i++] = '+';

	status[i] = '\0';
	sendto_one(sptr, rpl_str(RPL_WHOREPLY),
		me.name,
		sptr->name,
		(repchan) ? (repchan->chname) : "*",
		acptr->user->username,
		IsHMask(acptr) && !(su && show_real_host)
			? acptr->user->vhost : acptr->user->host,
		(BadPtr(server_mask) || (sptr == acptr) || su)
			? acptr->srvptr->name : server_mask,
		acptr->name,
		status,
		(BadPtr(server_mask) || IsOper(sptr)) ? acptr->hopcount : 1,
		acptr->info);
}


/* parv[0] = sender prefix, [1] = mask, parv[2] = flags */
int m_who(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	Link *lp;
	aClient *clptr, *fcl[MAXCHANNELSPERUSER];
	aChannel *chptr, *fch[MAXCHANNELSPERUSER];
	char mask[HOSTLEN+NICKLEN+USERLEN+3], *p = NULL, *par, buf[BUFSIZE];
	int i, j, member, clidx = 0, chidx = 0, chcnt = 0, umreq = 0,
		umden = 0, checkops[MAXCHANNELSPERUSER];

	if (check_registered_user(sptr))
		return 0;

	if (parc > 2) {
		char c = '+';
		for (i = 0; parv[2][i]; i++) {
			if (parv[2][i] == '+' || parv[2][i] == '-') {
				c = parv[2][i];
			} else {
				unsigned int f = 0;
				switch (parv[2][i]) {
					case 'o':
					case 'O':
						f |= UMODE_OPER|UMODE_LOCOP;
						break;
					case 'a':
						f |= UMODE_SADMIN;
						break;
					case 'A':
						f |= UMODE_ADMIN;
						break;
					case 'r':
						f |= UMODE_REGNICK;
						break;
					case 'w':
						if (IsAnOper(sptr))
							f |= UMODE_WALLOP;
						break;
					case 'i':
						f |= UMODE_INVISIBLE;
						break;
					case 'x':
						f |= UMODE_HMASK;
						break;
					case 'g':
					case 'G':
						f |= UMODE_GONE;
						break;
				}

				if (c == '+')
					umreq |= f;
				else
					umden |= f;
			}
		}
	}

	show_real_host = (parc > 3 && !strcasecmp(parv[3], "--real")) ? 1 : 0;

	mask[0] = '\0';
	mask[sizeof(mask)-1] = '\0';
	fcl[0] = (aClient *) NULL;
	fch[0] = (aChannel *) NULL;

	strncpyzt(buf, (parv[1] ? parv[1] : "*"), sizeof(buf));

	/* dispatch parameters */
	for (par = strtoken(&p, parv[1], ","); par; par = strtoken(&p, NULL, ",")) {
		/* channel name */
		char c = par[*par == '@' ? 1 : 0];
		if (c == '#' || c == '&' || c == '+') {
			/* make sure there's enough room for a new channel */
			if (chidx < elementsof(fch)) {
				if ((chptr = find_channel(*par == '@' ? par+1 : par, NullChn)) != NullChn) {
					if (IsMember(sptr, chptr) || (chptr->mode.mode & MODE_SECRET|MODE_PRIVATE) || IsAnOper(sptr)) {
						fch[chidx] = chptr;
						checkops[chidx++] = (*par == '@' ? 1 : 0);
					}
				}
			}

			/* count channel attempts, not successes - to prevent full 
			   listing when a cliend is trying to WHO an unexisting channel */
			++chcnt;
		}

		/* or nicks */
		else if ((clptr = find_client(par, (aClient *)NULL)) && IsPerson(clptr)) {
			if (clidx < elementsof(fcl))
				fcl[clidx++] = clptr;
		}

		/* or masks */
		else {
			strncpyzt(mask, par, sizeof(mask));
			collapse(mask);
		}
	}

	/* truncate lists */
	if (chidx < elementsof(fch)) fch[chidx] = (aChannel *) NULL;
	if (clidx < elementsof(fcl)) fcl[clidx] = (aClient *) NULL;

	/* backwards compatibility */
	if (mask[0] == '0' && !mask[1])
		mask[0] = '*';

	/* adjust marker so that it's never NULL (which is the default marker's
	   state in the client structures) */
	do { ++whox_mark; } while (!whox_mark);

	/* if we have channel names supplied, process them only */
	if (fch[0] || chcnt) {
		for (i = 0; i < elementsof(fch) && fch[i]; i++) {
			member = IsMember(sptr, fch[i]);

			for (lp = fch[i]->members; lp; lp = lp->next) {
				clptr = lp->value.cptr;

				if (!IsPerson(clptr) || IsULine(clptr,clptr))
					continue;

				if (checkops[i] && !(lp->flags & CHFL_CHANOP))
					continue;

				/* skip already processed clients */
				if (clptr->whox_mark == whox_mark)
					continue;

				/* filter flags */
				if ((umreq && !(clptr->umodes & umreq)) || (umden && (clptr->umodes & umden)))
					continue;

				/* if we have a list of clients, check that this one
				   if one of them */
				if (fcl[0]) {
					for (j = 0; fcl[j] && j < elementsof(fcl); j++)
						if (fcl[j] == clptr)
							break;
					if (j = elementsof(fcl) || !fcl[j])
						continue;
				}

				/* hide invisible users */
				if (clptr != sptr && IsInvisible(clptr) && !member && !IsOper(sptr) && !(IsLocOp(sptr) && MyClient(clptr)))
					continue;

				/* host matching routines are inside now */
				do_who(sptr, clptr, fch[i], mask);

				/* don't send this one again */
				clptr->whox_mark = whox_mark;
			}
		}
	}

	/* got nicks - process them only */
	else if (fcl[0]) {
		for (i = 0; i < elementsof(fcl) && (clptr = fcl[i]); i++) {
			if (!IsPerson(clptr) || IsULine(clptr,clptr))
				continue;

			if (clptr->whox_mark == whox_mark)
				continue;

			if ((umreq && !(clptr->umodes & umreq))
			|| (umden && (clptr->umodes & umden)))
				continue;

			do_who(sptr, clptr, NULL, mask);
		}
	}

	/* the worst case.  match agains all clients. */
	else {
		for (clptr = client; clptr; clptr = clptr->next) {
			if (!IsPerson(clptr) || IsULine(clptr,clptr))
				continue;

			if (clptr->whox_mark == whox_mark)
				continue;

			if ((umreq && !(clptr->umodes & umreq))
			|| (umden && (clptr->umodes & umden)))
				continue;

			do_who(sptr, clptr, NULL, mask);
		}
	}

	sendto_one(sptr, rpl_str(RPL_ENDOFWHO), me.name, parv[0], buf);
	return 0;
}
