/* parse.c -- parse received data and call the appropriate command handlers;
 * Copyright (C) 2001-2002, ForestNet IRC Network
 * This is a free software; see file `license' for details.
 * $Id: parse.c,v 1.19 2002/11/18 12:44:08 hex Exp $
 */

#include <string.h>

#include "struct.h"
#include "common.h"
#define MSGTAB
#include "msg.h"
#undef MSGTAB
#include "sys.h"
#include "numeric.h"
#include "h.h"
#include "unicode.h"
#include "send.h"

#ifndef MIN
# define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

/* NOTE: parse() should not be called recursively by other functions! */
static char *para[MAXPARA + 1];

static char sender[HOSTLEN + 1];


static int cancel_clients(aClient * cptr, aClient * sptr, char *cmd)
{
	if (IsServer(sptr) || IsMe(sptr) || IsServer(cptr))
		return 0;

	return exit_client(cptr, cptr, &me, "Fake prefix");
}


static void remove_unknown(aClient * cptr, char *sender)
{
	if (!IsRegistered(cptr) || IsClient(cptr) || !IsServer(cptr))
		return;

	/* Do kill if it came from a server because it means there is a ghost
	 * user on the other server which needs to be removed. -avalon */
	if (!strchr(sender, '.')) {
		if (BadPtr(server_mask))
			sendto_one(cptr, ":%s KILL %s :%s (%s(?) <- %s)",
				me.name, sender, me.name, sender,
				get_client_name(cptr, FALSE));
		else
			sendto_one(cptr, ":%s KILL %s :Ghost", me.name, sender);
	}
	else
		sendto_one(cptr, ":%s SQUIT %s :(Unknown from %s)", me.name,
			sender, get_client_name(cptr, FALSE));
}


/*  Find a client (server or user) by name.
**
**  *Note*
**     Semantics of this function has been changed from
**     the old. 'name' is now assumed to be a null terminated
**     string and the search is the for server and user.
*/
aClient *find_client(char *name, aClient * cptr)
{
	return name ? hash_find_client(name, cptr) : cptr;
}


/*
**  Find server by name.
**
**     This implementation assumes that server and user names
**     are unique, no user can have a server name and vice versa.
**     One should maintain separate lists for users and servers,
**     if this restriction is removed.
**
**  *Note*
**     Semantics of this function has been changed from
**     the old. 'name' is now assumed to be a null terminated
**     string.
*/
aClient *find_server(char *name, aClient * cptr)
{
	return name ? hash_find_server(name, cptr) : cptr;
}


aClient *find_name(char *name, aClient * cptr)
{
	aClient *c2ptr = cptr;

	if (!collapse(name))
		return c2ptr;

	if ((c2ptr = hash_find_server(name, cptr)))
		return (c2ptr);

	if (!strchr(name, '*'))
		return c2ptr;

	for (c2ptr = client; c2ptr; c2ptr = c2ptr->next) {
		if (!IsServer(c2ptr) && !IsMe(c2ptr))
			continue;
		if (match(name, c2ptr->name) == 0)
			break;
		if (strchr(c2ptr->name, '*'))
			if (match(c2ptr->name, name) == 0)
				break;
	}

	return c2ptr ? c2ptr : cptr;
}

/* Find person by (nick)name. */
aClient *find_person(char *name, aClient *cptr)
{
	aClient *c2ptr = find_client(name, cptr);

	return (c2ptr && IsClient(c2ptr) && c2ptr->user)
		? c2ptr : cptr;
}


/*
 * parse a buffer.
 *
 * NOTE: parse() should not be called recusively by any other fucntions!
 */
int parse(aClient *cptr, char *buffer, char *bufend, struct Message *mptr)
{
	aClient *from = cptr;
	char *ch, *s, parsebuf[MAXMSGLEN+1];
	int len, i, numeric, paramcount, noprefix = 0;
#ifdef DEBUGMODE
	time_t then, ticks;
	int retval;
#endif

	if (IsDead(cptr))
		return 0;

	if (IsClient(cptr) && cptr->charset != NULL) {
		/* copy the message */
		size_t size = MIN(bufend - buffer, MAXMSGLEN);
		strncpy(parsebuf, buffer, size);
		parsebuf[size] = '\0';

		/* translate it */
		charset_convertfrom(parsebuf, sizeof(parsebuf), cptr->charset);

		/* re-assign the pointers */
		buffer = parsebuf;
		bufend = parsebuf + strlen(parsebuf) - 1;
	}

	*(s = sender) = '\0';
	for (ch = buffer; *ch == ' '; ++ch);
	para[0] = from->name;

	if (*ch == ':') {
		/* Copy the prefix to 'sender' assuming it terminates with
		 * SPACE (or NULL, which is an error, though). */
		for (++ch, i = 0; *ch && *ch != ' '; ch++)
			if (s < (sender + sizeof(sender) - 1))
				*s++ = *ch; /* leave room for NULL */
		*s = '\0';
		/* Actually, only messages coming from servers can have the
		   prefix--prefix silently ignored, if coming from a user client...
		   ...sigh, the current release "v2.2PL1" generates also null prefixes,
		   at least to NOTIFY messages (e.g. it puts "sptr->nickname" as prefix
		   from server structures where it's null--the following will handle
		   this case as "no prefix" at all --msa (": NOTICE nick ...") */
		if (*sender && IsServer(cptr)) {
			from = find_client(sender, (aClient *) NULL);
			if (!from || match(from->name, sender))
				from = find_server(sender, (aClient *) NULL);

			para[0] = sender;

			/* Hmm! If the client corresponding to the prefix is not
			   found--what is the correct action??? Now, I will ignore the
			   message (old IRC just let it through as if the prefix just
			   wasn't there...) --msa */
			if (!from) {
				Debug((DEBUG_ERROR, "Unknown prefix (%s)(%s) from (%s)", sender, buffer, cptr->name));
				ircstp->is_unpf++;
				remove_unknown(cptr, sender);
				return -1;
			}
			if (from->from != cptr) {
				ircstp->is_wrdi++;
				Debug((DEBUG_ERROR, "Message (%s) coming from (%s)", buffer, cptr->name));
				return cancel_clients(cptr, from, ch);
			}
		}
		while (*ch == ' ')
			ch++;
	}
	else
		noprefix = 1;

	if (*ch == '\0') {
		ircstp->is_empt++;
		Debug((DEBUG_NOTICE, "Empty message from host %s:%s", cptr->name, from->name));
		return (-1);
	}

	/* Extract the command code from the packet.  Point s to the end of
	   the command code and calculate the length using pointer arithmetic.
	   Note: only need length for numerics and *all* numerics must have
	   paramters and thus a space after the command code. -avalon */
	s = strchr(ch, ' '); /* s -> End of the command code */
	len = (s) ? (s - ch) : 0;
	if (len == 3 && isdigit(*ch) && isdigit(*(ch + 1)) && isdigit(*(ch + 2))) {
		mptr = NULL;
		numeric = (*ch - '0') * 100 + (*(ch + 1) - '0') * 10 + (*(ch + 2) - '0');
		paramcount = MAXPARA;
		ircstp->is_num++;
	}
	else {
		if (s)
			*s++ = '\0';
		if (ch[1] == '\0' && IsToken(cptr))
			mptr = msgmap[(u_char) * ch];
		else
			for (; mptr->cmd; mptr++)
				if (mycmp(mptr->cmd, ch) == 0)
					break;

		if (!mptr->cmd) {
			/* Note: Give error message *only* to recognized persons.  It's a
			   nightmare situation to have two programs sending "Unknown
			   command"'s or equivalent to each other at full blast....  If it
			   has got to person state, it at least seems to be well behaving.
			   Perhaps this message should never be generated, though...  --msa
			   Hm, when is the buffer empty -- if a command code has been
			   found ?? -Armin */
			if (buffer[0] != '\0') {
				if (IsPerson(from))
					sendto_one(from, ":%s %d %s %s :Unknown command", me.name, ERR_UNKNOWNCOMMAND, from->name, ch);
				Debug((DEBUG_ERROR, "Unknown (%s) from %s", ch, get_client_name(cptr, TRUE)));
			}
			ircstp->is_unco++;
			return (-1);
		}
		paramcount = mptr->parameters;
		i = bufend - ch; /* Is this right? -Donwulff */
		mptr->bytes += i;
		if ((mptr->flags & 1) && !(IsServer(cptr) || IsAnOper(cptr)))
			cptr->since += (2 + i / 120);
		/* Allow only 1 msg per 2 seconds (on average) to prevent
		   dumping. to keep the response rate up, bursts of up
		   to 5 msgs are allowed -SRB */
	}

	i = 0;
	if (s) {
		if (paramcount > MAXPARA)
			paramcount = MAXPARA;
		for (;;) {
			while (*s == ' ')
				*s++ = '\0';

			if (*s == '\0')
				break;
			if (*s == ':') {
				para[++i] = s + 1;
				break;
			}
			para[++i] = s;
			if (i >= paramcount)
				break;
			for (; *s != ' ' && *s; s++);
		}
	}
	para[++i] = NULL;
	if (mptr == NULL)
		return (do_numeric(numeric, cptr, from, i, para));
	mptr->count++;

	if (IsRegisteredUser(cptr) &&
#ifdef IDLE_FROM_MSG
		mptr->func == m_private)
#else
		mptr->func != m_ping && mptr->func != m_pong)
#endif
		from->user->last = time(NULL);

#ifndef DEBUGMODE
	return (*mptr->func) (cptr, from, i, para);
#else
	then = clock();
	retval = (*mptr->func) (cptr, from, i, para);
	if (retval != FLUSH_BUFFER) {
		ticks = (clock() - then);
		if (IsServer(cptr))
			mptr->rticks += ticks;
		else
			mptr->lticks += ticks;
		cptr->cputime += ticks;
	}

	return retval;
#endif
}
